/*
 *	MikDLL - Done by MikMak / HaRDCoDE '95
 *	contributions - Paul Hsieh
 */
#include <string.h>
#include "mdll.h"
#include "mdllP.h"

struct mdllInit memApi = { (void *)0, (void *)0 };

int MDLL_Init (struct mdllInit * initCfg) {
    if (!initCfg) return -__LINE__;
    memApi = *initCfg;
    return 0;
}

#define chkSanity(retval) { if (mdllH == (void *)0 || memApi.malloc == (void *)0 || memApi.free == (void *)0) return (retval); }

mdllHandle MDLL_NewInterface (void) {
mdllHandle mdllH;
mdllHandle mdllAltH;

    if (memApi.malloc == (void *)0 || memApi.free == (void *)0) return (void *)0;
    mdllH = (mdllHandle)memApi.malloc (2 * sizeof (struct dllIntfH));
    if (mdllH) {
        mdllAltH = &mdllH[1];

        mdllH->isParent = 1;
        mdllH->alt = mdllAltH;
        mdllH->pefile = (void *)0;
        mdllH->pool = (void *)0;
        mdllH->dll_error = (void *)0;
        mdllH->msgbuf[0] = '\0';

        mdllAltH->isParent = 0;
        mdllAltH->alt = mdllH;
        mdllAltH->pefile = (void *)0;
        mdllAltH->pool = (void *)0;
        mdllAltH->dll_error = (void *)0;
        mdllAltH->msgbuf[0] = '\0';
    }
    return mdllH;
}

int MDLL_DestroyInterface (mdllHandle mdllH) {

    chkSanity (-1);

    if (mdllH->isParent == 1) {
	int ret = MDLL_DestroyInterface (mdllH->alt);
	if (ret) return ret;
        mdllH->alt = NULL;
    } else if (mdllH->isParent != 0) return -__LINE__;

    if (mdllH->pefile) MDLL_Unbind (mdllH);
    mdllH->pefile = (void *)0;
    if (mdllH->pool) MDLL_RemoveExport (mdllH, (void *)0);
    mdllH->pool = (void *)0;
    mdllH->msgbuf[0] = '\0';
    mdllH->dll_error = "Erroneous access, this object is destroyed";

    if (mdllH->isParent == 1) {
        mdllH->isParent = -1;
        memApi.free (mdllH);
    } else {
        mdllH->isParent = -1;
    }

    return 0;
}

char * MDLL_GetErr (mdllHandle mdllH) {
    chkSanity ((void *)0);
    return mdllH->dll_error;
}

static EXPORTENTRY *findexportentry (mdllHandle mdllH, char *name) {
	EXPORTENTRY *e;

        chkSanity ((void *)0);

	// crit-section (mdllH) {
	e = mdllH->pool;
	while (e != (void *)0) {
		if (!strcmp (e->name, name)) break;
		e = e->next;
	}
	// }

	return e;
}


int MDLL_WalkPool (mdllHandle mdllH, int (* iter)(void * parm, char * name), void * parm) {
	EXPORTENTRY *e;

	chkSanity (-__LINE__);
	if (iter  == (void *)0) return -__LINE__;
	mdllH = mdllH->alt;
	chkSanity (-__LINE__);

	// crit-section (mdllH) {
	e = mdllH->pool;
	while (e != (void *)0) {
		int ret = iter (parm, e->name);
		if (ret) return ret;
		e = e->next;
	}
	// }

	return 0;
}



int MDLL_Export (mdllHandle mdllH, char *name, void *obj) {
	EXPORTENTRY *e;

	chkSanity (-__LINE__);

	// see if a export entry by this name already exists
	if (findexportentry (mdllH, name) != (void *)0) return __LINE__;

	if ((e = memApi.malloc (sizeof (EXPORTENTRY))) != (void *)0){
		e->name = name;
		e->obj  = obj;
		// crit-section (mdllH) {
		e->next = mdllH->pool;
		mdllH->pool = e;
		// }
	} else return -__LINE__;

	return 0;
}



int MDLL_RemoveExport (mdllHandle mdllH, char *name) {
	EXPORTENTRY *e, *o;

	chkSanity (-__LINE__);

	if (name == (void *)0) {
		int ret = 0;
		e = mdllH->pool;
		while (e) {
			if ((ret = MDLL_RemoveExport (mdllH, e->name)) != 0)
				break;
			e = mdllH->pool;
		}
		return ret;
	}

	if ((o = findexportentry (mdllH, name)) == (void *)0) return __LINE__;

	/* Copy over found entry with first entry, then remove first entry */
	e = mdllH->pool;
	o->name = e->name;
	o->obj  = e->obj;
	mdllH->pool = e->next;
	memApi.free (e);
	return 0;
}



void *MDLL_Import (mdllHandle mdllH, char *name) {
	EXPORTENTRY *e;

	chkSanity ((void *)0);
	if (mdllH->dll_error != (void *)0) return (void *)0;
	mdllH = mdllH->alt;
	chkSanity ((void *)0);
	if (mdllH->dll_error != (void *)0) return (void *)0;

	if ((e = findexportentry (mdllH, name)) == (void *)0){
		/* BOZO -- buffer overflow here */
		strcpy  (mdllH->msgbuf, "Couldn't resolve external reference: '");
		strncat (mdllH->msgbuf, name, 96);
		strcat  (mdllH->msgbuf, "'");
		return (void *)0;
	}

	return e->obj;
}


int MDLL_Unbind (mdllHandle mdllH) {
	chkSanity (-__LINE__);

	/* Parent can call this, child cannot. */
	if (mdllH->isParent != 1) return -__LINE__;
	if (mdllH->pefile == (void *)0) return -__LINE__;

	/* Wipe out all imports from the DLL */
	MDLL_RemoveExport (mdllH->alt, (void *)0);

	PE_Free (mdllH->pefile);
	mdllH->pefile = (void *)0;
	return 0;
}


int MDLL_Bind (mdllHandle mdllH, char *name) {

	PE *pefile;
	int (*entryp) (void *(*MDLL_Import)(), void (*MDLL_Export)(), mdllHandle);

	chkSanity (-__LINE__);
	if (mdllH->isParent != 1) return -__LINE__;

	// reset error condition
	mdllH->dll_error = (void *)0;

	// try to read the file (portable executable format)
	if ((pefile = PE_ReadFN (mdllH, name)) == (void *)0) return -__LINE__;

	// find the entrypoint
	entryp = PE_ImportFnc (pefile, "mdll_entry_");
	if (entryp == (void *)0){
		entryp = PE_ImportFnc (pefile, "_mdll_entry");
	}

	if (entryp == (void *)0) {
		PE_Free (pefile);
		mdllH->dll_error = "Couldn't find MDLL entrypoint";
		return -__LINE__;
	}

	// execute entrypoint
	entryp (MDLL_Import, MDLL_Export, mdllH->alt);

	// check if entrypoint was successfull
	if (mdllH->dll_error != (void *)0){
		// no ? free pefile and return (void *)0;
		PE_Free (pefile);
		return -__LINE__;
	}

	mdllH->pefile = pefile;
	return 0;
}
